// Copyright (c) 2015-present Mattermost, Inc. All Rights Reserved.
// See LICENSE.txt for license information.
import React, {useContext, useState} from 'react'
import {FormattedMessage, useIntl} from 'react-intl'

import {Archiver} from '../../archiver'
import {darkTheme, defaultTheme, lightTheme, setTheme, Theme} from '../../theme'
import Menu from '../../widgets/menu'
import MenuWrapper from '../../widgets/menuWrapper'
import {SetLanguageContext} from '../../setLanguageContext'
import {UserSettings} from '../../userSettings'

import './sidebarSettingsMenu.scss'

type Props = {
    setWhiteLogo: (whiteLogo: boolean) => void
}

const SidebarSettingsMenu = React.memo((props: Props) => {
    const intl = useIntl()
    const setLanguage = useContext(SetLanguageContext)

    const updateTheme = (theme: Theme | null) => {
        const consolidatedTheme = setTheme(theme)
        const whiteLogo = (consolidatedTheme.sidebarWhiteLogo === 'true')
        props.setWhiteLogo(whiteLogo)
    }

    const [randomIcons, setRandomIcons] = useState(UserSettings.prefillRandomIcons)
    const toggleRandomIcons = () => {
        UserSettings.prefillRandomIcons = !UserSettings.prefillRandomIcons
        setRandomIcons(!randomIcons)
    }

    return (
        <div className='SidebarSettingsMenu'>
            <MenuWrapper>
                <div className='menu-entry'>
                    <FormattedMessage
                        id='Sidebar.settings'
                        defaultMessage='Settings'
                    />
                </div>
                <Menu position='top'>
                    <Menu.Text
                        id='import'
                        name={intl.formatMessage({id: 'Sidebar.import-archive', defaultMessage: 'Import archive'})}
                        onClick={async () => Archiver.importFullArchive()}
                    />
                    <Menu.Text
                        id='export'
                        name={intl.formatMessage({id: 'Sidebar.export-archive', defaultMessage: 'Export archive'})}
                        onClick={async () => Archiver.exportFullArchive()}
                    />
                    <Menu.SubMenu
                        id='lang'
                        name={intl.formatMessage({id: 'Sidebar.set-language', defaultMessage: 'Set language'})}
                        position='top'
                    >
                        <Menu.Text
                            id='english-lang'
                            name={intl.formatMessage({id: 'Sidebar.english', defaultMessage: 'English'})}
                            onClick={async () => setLanguage('en')}
                        />
                        <Menu.Text
                            id='spanish-lang'
                            name={intl.formatMessage({id: 'Sidebar.spanish', defaultMessage: 'Spanish'})}
                            onClick={async () => setLanguage('es')}
                        />
                        <Menu.Text
                            id='german-lang'
                            name={intl.formatMessage({id: 'Sidebar.german', defaultMessage: 'German'})}
                            onClick={async () => setLanguage('de')}
                        />
                        <Menu.Text
                            id='japanese-lang'
                            name={intl.formatMessage({id: 'Sidebar.japanese', defaultMessage: 'Japanese'})}
                            onClick={async () => setLanguage('ja')}
                        />
                        <Menu.Text
                            id='french-lang'
                            name={intl.formatMessage({id: 'Sidebar.french', defaultMessage: 'French'})}
                            onClick={async () => setLanguage('fr')}
                        />
                        <Menu.Text
                            id='dutch-lang'
                            name={intl.formatMessage({id: 'Sidebar.dutch', defaultMessage: 'Dutch'})}
                            onClick={async () => setLanguage('nl')}
                        />
                        <Menu.Text
                            id='russian-lang'
                            name={intl.formatMessage({id: 'Sidebar.russian', defaultMessage: 'Russian'})}
                            onClick={async () => setLanguage('ru')}
                        />
                        <Menu.Text
                            id='chinese-lang'
                            name={intl.formatMessage({id: 'Sidebar.chinese', defaultMessage: 'Traditional Chinese'})}
                            onClick={async () => setLanguage('zh')}
                        />
                        <Menu.Text
                            id='simplified-chinese-lang'
                            name={intl.formatMessage({id: 'Sidebar.simplified-chinese', defaultMessage: 'Simplified Chinese'})}
                            onClick={async () => setLanguage('zh_Hans')}
                        />
                        <Menu.Text
                            id='turkish-lang'
                            name={intl.formatMessage({id: 'Sidebar.turkish', defaultMessage: 'Turkish'})}
                            onClick={async () => setLanguage('tr')}
                        />
                        <Menu.Text
                            id='occitan-lang'
                            name={intl.formatMessage({id: 'Sidebar.occitan', defaultMessage: 'Occitan'})}
                            onClick={async () => setLanguage('oc')}
                        />
                    </Menu.SubMenu>
                    <Menu.SubMenu
                        id='theme'
                        name={intl.formatMessage({id: 'Sidebar.set-theme', defaultMessage: 'Set theme'})}
                        position='top'
                    >
                        <Menu.Text
                            id='default-theme'
                            name={intl.formatMessage({id: 'Sidebar.default-theme', defaultMessage: 'Default theme'})}
                            onClick={async () => updateTheme(defaultTheme)}
                        />
                        <Menu.Text
                            id='dark-theme'
                            name={intl.formatMessage({id: 'Sidebar.dark-theme', defaultMessage: 'Dark theme'})}
                            onClick={async () => updateTheme(darkTheme)}
                        />
                        <Menu.Text
                            id='light-theme'
                            name={intl.formatMessage({id: 'Sidebar.light-theme', defaultMessage: 'Light theme'})}
                            onClick={async () => updateTheme(lightTheme)}
                        />
                        <Menu.Text
                            id='system-theme'
                            name={intl.formatMessage({id: 'Sidebar.system-theme', defaultMessage: 'System theme'})}
                            onClick={async () => updateTheme(null)}
                        />
                    </Menu.SubMenu>
                    <Menu.Switch
                        id='random-icons'
                        name={intl.formatMessage({id: 'Sidebar.random-icons', defaultMessage: 'Random icons'})}
                        isOn={randomIcons}
                        onClick={async () => toggleRandomIcons()}
                    />
                </Menu>
            </MenuWrapper>
        </div>
    )
})

export default SidebarSettingsMenu
